import { chain, curry } from 'izuna'
import { ok } from 'kojima'
import { clone, diff, fail, succeed } from './fn.js'
import { state as State } from './state.js'
import { skip } from './cond.js'

/** @import { ParserState } from './state.js' */

/**
 * @param {...Parser} parsers
 */
export const seq = (...parsers) =>
  /** @param {ParserState} state */
  state => {
    let acc = ok(state)
    for (const parser of parsers) {
      if (acc.isOk()) {
        acc = chain(parser, acc)
      } else {
        break
      }
    }

    return acc
  }

/**
 * @param {...any} parsers
 */
export const any = (...parsers) =>
  /**
   * @param {ParserState} state
   */
  state => {
    for (const parser of parsers) {
      //console.log('combinator.js::35/any:', state)
      const result = parser(clone(state))
      if (result.isOk()) {
        return result
      }
    }

    return fail('no matching parsers', state)
  }

export const map = curry(
  /** 
   * @param {(...args: any[]) => any} fn
   * @param {Parser} parser
   * @param {ParserState} state
   */
  (fn, parser, state) => {
    return parser(state).chain(result => {
      try {
        /** @type {Result<ParserState, ParseError>} */
        const parsed = fn(diff(state[0], result[0]))
        const backtrack = State(result[1], state[0])

        return succeed(parsed, backtrack)
      } catch (e) {
        return fail('failed to map', state, e)
      }
    })
  })

export const delimited = curry((first, second, third, state) => seq(skip(first), second, skip(third))(state))
export const preceded = curry((first, second, state) => seq(skip(first), second)(state))
export const terminated = curry((first, second, state) => seq(first, skip(second))(state))

export const separatedPair = curry((first, delimiter, second, state) => seq(first, skip(delimiter), second)(state))

