import { ok } from 'kojima'
import { chain, compose, curry, pipe } from 'izuna'
import { eof } from './byte.js'
import { Tuple } from './tuple.js'
import { iter } from './fn.js'
import { Clone } from './clone.js'

/**
 * @typedef {Readonly<[any[], Iterator<any>]>} ParserState
 */

export class ParseError extends Error {
  /**
   * @param {string} message
   * @param {ParserState} state
   * @param {Error} [cause]
   */
  constructor(message, state, cause) {
    super(message, { cause })
    this.state = state
  }
}

export class State extends Tuple {
  constructor(remaining, read = []) {
    super([['read', read], ['remaining', remaining]])
  }

  static from(values) {
    return new State(iter(values))
  }

  next() {
    return this.remaining.next()
  }

  toString() {
    return `State(${this.read}, ${[...this.clone().remaining]})`
  }

  [Clone]() {
    return Object.freeze(new State(this.remaining.clone(), this.read))
  }

  [Symbol.iterator]() {
    return super[Symbol.iterator]()
  }
}

export const state = (remaining, read) => new State(remaining, read)

export const parse = curry((parser, input) => pipe(
  iter,
  state,
  ok,
  chain(parser),
)(input))

export const parseAll = curry((parser, input) => compose(eof, parse(parser), input))

